!---------------------------------------------------------------------------------------
! M1QN3 algorithm and supporting routines
!---------------------------------------------------------------------------------------
!
!     M1QN3, Version 3.3, October 2009
!
!     M1qn3 has two running modes: the SID (Scalar Initial Scaling) mode
!     and the DIS (Diagonal Initial Scaling) mode. Both do not require
!     the same amount of storage, the same subroutines, ...
!     In the description below, items that differ in the DIS mode with
!     respect to the SIS mode are given in brakets.
!
!     Use the following subroutines:
!         M1QN3A
!         DD, DDS
!         NLIS0 + ECUBE (Dec 88)
!         MUPDTS, YSTBL.
!
!     The following routines are proposed to the user in case the
!     Euclidean scalar product is used:
!         EUCLID, CTONB, CTCAB.
!
!     La sous-routine M1QN3 est une interface entre le programme
!     appelant et la sous-routine M1QN3A, le minimiseur proprement dit.
!
!     Le module PROSCA est sense realiser le produit scalaire de deux
!     vecteurs de Rn; le module DTONB est sense realiser le changement
!     de coordonnees correspondant au changement de bases: base
!     euclidienne -> base orthonormale (pour le produit scalaire
!     PROSCA); le module CTBAB fait la transformation inverse: base
!     orthonormale -> base euclidienne.
!
!     Iz is an integer working zone for M1QN3A, its dimension is 5.
!     It is formed of 5 scalars that are set by the optimizer:
!         - the dimension of the problem,
!         - an identifier of the scaling mode,
!         - the number of updates,
!         - two pointers.
!
!     Dz est la zone de travail pour M1QN3A, de dimension ndz.
!     Elle est subdivisee en
!         3 [ou 4] vecteurs de dimension n: d,gg,[diag,]aux
!         m scalaires: alpha
!         m vecteurs de dimension n: ybar
!         m vecteurs de dimension n: sbar
!
!     m est alors le plus grand entier tel que
!         m*(2*n+1)+3*n .le. ndz [m*(2*n+1)+4*n .le. ndz)]
!     soit m := (ndz-3*n) / (2*n+1) [m := (ndz-4*n) / (2*n+1)].
!     Il faut avoir m >= 1, donc ndz >= 5n+1 [ndz >= 6n+1].
!
!     A chaque iteration la metrique est formee a partir d un multiple
!     de l'identite [d'une matrice diagonale] D qui est mise a jour m
!     fois par la formule de BFGS en utilisant les m couples {y,s} les
!     plus recents.
!
!     Authors: Jean Charles Gilbert, Claude Lemarechal, INRIA.
!
!     Copyright 2008, 2009, INRIA.
!
!     M1QN3 is distributed under the terms of the GNU General Public
!     License.
!
!     Rewritten in fortran90 by Rona Thompson, NILU, 2018
!
!---------------------------------------------------------------------------------------
!
!     This program is free software: you can redistribute it and/or
!     modify it under the terms of the GNU General Public License as
!     published by the Free Software Foundation, either version 3 of
!     the License, or (at your option) any later version.
!
!     This program is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
!     General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with this program.  If not, see
!     <http://www.gnu.org/licenses/>.
!
!---------------------------------------------------------------------------------------

module mod_m1qn3

  implicit none

  contains

  !------------------------------------------------------------
  ! M1QN3
  !------------------------------------------------------------

      subroutine m1qn3 (n,x,f,g,dxmin,df1,&
                        epsg,normtype,impres,io,imode,omode,niter,nsim,&
                        iz,dz,ndz,reverse,indic,isim,dzs,&
                        inmemo,sscale,ntravu,id,igg,idiag,iaux,ialpha,iybar,isbar,m,&
                        mmemo,reentry,gnorm,gnorms,ps,dk,dk1,&
                        cold,warm,skip_update,&
                        itmax,moderl,jcour,&
                        d1,t,tmin,tmax,eps1,ff,preco,precos,ys,den,ps2,hp0,&
                        t_increased,indica,indicd,&
                        tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd,&
                        fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa)

      implicit none

    ! arguments

      character(len=3) :: normtype
      integer :: n,impres,io,imode(3),omode,niter,nsim,iz(5),ndz,indic,reverse
      real(kind=8) :: x(n),f,g(n),dxmin,df1,epsg,dz(ndz),dzs(*)

    ! local variables

    !  The variable 'reentry' is used to know where to jump when entering
    !  the subroutine in reverse commnunication (it is better than using
    !  'reverse', which is known outside m1qn3 and could be changed by
    !  the user):
    !  = 0: no jump, start at the begining of the subroutine
    !  = 1: jump inside m1qn3a, where the simulator was called with
    !       indic=1
    !  = 2: jump inside m1qn3a/mlis3, where the simulator was called with
    !       indic=4
    !  The variable 'reentry' is set to 0 by the compiler and when m1qn3
    !  has completed its job (see below); since the value is saved, it
    !  has always the value 0 when entering the solver for the first
    !  time when solving a new problem, even if it is in the same
    !  program.

      logical :: inmemo,sscale
      integer :: ntravu,id,igg,idiag,iaux,ialpha,iybar,isbar,m,mmemo,reentry
      real(kind=8) :: gnorm,ps

    ! variables defined in m1qn3a but need to be saved

      logical :: cold,warm,skip_update
      integer :: i,itmax,moderl,isim,jcour
      real(kind=8) :: d1,t,tmin,tmax,gnorms,eps1,ff,preco,precos,ys,den,dk,dk1,ps2,hp0

    ! variables defined in mlis3 but need to be saved

      logical :: t_increased
      integer :: indica,indicd
      real(kind=8) :: tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd
      real(kind=8) :: fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa

    ! stop if reverse < 0 (m1qn3 should not be called with reverse < 0)

      if (reverse.lt.0) then
          write (io,'(/a,a,i0,a/)') &
             " >>> m1qn3 should not be called with a negative reverse", &
             " (=", reverse, ")"
          stop
      endif

    ! possible jumps
    !  9999: go directly in m1qn3a

      if (reentry.gt.0) goto 9999

    ! license notice

      if (impres.ge.5) then
          write (io,934)
      endif
  934 format (1x,79("-") &
         /1x,"M1QN3 Copyright (C) 2008, J. Ch. Gilbert, Cl. ",&
             "Lemarechal."&
         /1x,79("-")&
         /1x,"This program comes with ABSOLUTELY NO WARRANTY. This is",&
             " free software, and you"&
         /1x,"are welcome to redistribute it under certain ",&
             "conditions. See the file COPYING "&
         /1x,"in the root directory of the M1QN3 distribution for ",&
             "details."&
         /1x,79("-"))

    ! impressions initiales et controle des arguments

      if (impres.ge.1) then
          write (io,900) n,dxmin,df1,epsg,normtype,niter,nsim,impres
          if (reverse.gt.0) then
              write (io,'(5x,a)') "reverse communication"
          else
              write (io,'(5x,a)') "direct communication"
          endif
      endif
  900 format (/" M1QN3 (Version 3.3, October 2009): entry point"/&
         5x,"dimension of the problem (n):",i14/&
         5x,"absolute precision on x (dxmin):",9x,1pd9.2/&
         5x,"expected decrease for f (df1):",11x,1pd9.2/&
         5x,"relative precision on g (epsg):",10x,1pd9.2,&
            " (",a3,"-norm)"/&
         5x,"maximal number of iterations (niter):",i6/&
         5x,"maximal number of simulations (nsim):",i6/&
         5x,"printing level (impres):",15x,i4)

    ! check the arguments

      if (n.le.0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: n should be > 0"
          return
      endif
      if (niter.le.0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: niter should be > 0"
          return
      endif
      if (nsim.le.0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: nsim should be > 0"
          return
      endif
      if (dxmin.le.0.d0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: dxmin should be > 0.d0"
          return
      endif
      if (epsg.le.0.d0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: epsg should be > 0.d0"
          return
      endif
      if (epsg.ge.1.d0) then
          omode=1
          niter=0
          nsim=0
          epsg=1.d0
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,'(/a)') &
             " >>> m1qn3: epsg is >= 1.d0, no need to make progress"
          goto 1000
      endif
      if ((normtype.ne.'two') .and. &
         (normtype.ne.'sup') .and. &
         (normtype.ne.'dfn')) then 
          omode=2
          if (reverse.gt.0) reverse = -1
          write (io,'(/a,a,a/)') " >>> m1qn3: unknown norm type '", &
             normtype, "'"
          return
      endif
      if (impres.lt.0) then
          omode=2
          if (reverse.gt.0) reverse = -1
          write (io,'(/a,i0/)') &
             " >>> m1qn3: impres should be >= 0 and has the value ", &
             impres
          return
      endif

    !  what method

      if (imode(1).eq.0) then
          if (impres.ge.1) write (io,920)
  920     format (/" m1qn3: Diagonal Initial Scaling mode")
          sscale=.false.
      else
          if (impres.ge.1) write (io,921)
  921     format (/" m1qn3: Scalar Initial Scaling mode")
          sscale=.true.
      endif

      if ((ndz.lt.5*n+1).or.((.not.sscale).and.(ndz.lt.6*n+1))) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,922)
  922     format (/" >>> m1qn3: not enough memory allocated")
          return
      endif

    ! Compute m

      call mupdts (sscale,inmemo,n,m,ndz)

      print*, 'm1qn3: m = ',m

    ! Check the value of m (if (y,s) pairs in core, m will be >= 1)

      if (m.lt.1) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,930)
  930     format (/" >>> m1qn3: m is set too small in mupdts")
          return
      endif

    ! mmemo = number of (y,s) pairs in core memory

      mmemo=1
      if (inmemo) mmemo=m

      ntravu=2*(2+mmemo)*n+m
      if (sscale) ntravu=ntravu-n
      if (impres.ge.1) write (io,931) ndz,ntravu,m
  931 format (/5x,"allocated memory (ndz) :",i9/ &
              5x,"used memory :           ",i9/ &
              5x,"number of updates :     ",i9)
      if (ndz.lt.ntravu) then
          omode=2
          if (reverse.gt.0) reverse = -1
          if (impres.ge.1) write (io,922)
          return
      endif

      if (impres.ge.1) then
          if (inmemo) then
              write (io,932)
          else
              write (io,933)
          endif
      endif
  932 format (5x,"(y,s) pairs are stored in core memory")
  933 format (5x,"(y,s) pairs are stored by the user")

    ! cold start or warm restart ?
    ! check iz: iz(1)=n, iz(2)=(0 if DIS, 1 if SIS),
    !           iz(3)=m, iz(4)=jmin, iz(5)=jmax

      if (imode(2).eq.0) then
          if (impres.ge.1) write (io,940)
      else
          if (iz(1).ne.n .or. iz(2).ne.imode(1) .or. iz(3).ne.m     .or. &
             iz(4).lt.1 .or. iz(5).lt.0        .or. iz(4).gt.iz(3) .or. &
             iz(5).gt.iz(3)) &
         then
              omode=2
              if (reverse.gt.0) reverse = -1
              if (impres.ge.1) then
                  write (io,941)
                  if (iz(1).ne.n) write (io,942)
                  if (iz(2).ne.imode(1)) write (io,943)
                  if (iz(3).ne.m) write (io,944)
                  if (iz(4).lt.1 .or. iz(5).lt.0 .or. iz(4).gt.iz(3) &
                     .or. iz(5).gt.iz(3)) write (io,945)
              endif
              return
          endif
          if (impres.ge.1) write (io,946)
      endif
  940 format (/" m1qn3: cold start"/1x)
  941 format (/" >>> m1qn3: inconsistent warm restart ")
  942 format (" >>> m1qn3: (the number of variables has changed)")
  943 format (" >>> m1qn3: (the scaling mode has changed)")
  944 format (" >>> m1qn3: (the number of updates has changed)")
  945 format (" >>> m1qn3: (wrong pointers)")
  946 format (/" m1qn3: warm restart"/1x)
      iz(1)=n
      iz(2)=0
      if (sscale) iz(2)=1
      iz(3)=m

    ! split the working zone dz

      idiag=1
      iybar=idiag+n
      if (sscale) iybar=1
      isbar=iybar+n*mmemo
      id=isbar+n*mmemo
      igg=id+n
      iaux=igg+n
      ialpha=iaux+n

    print*, 'm1qn3: id = ',id
    print*, 'm1qn3: igg = ',igg
    print*, 'm1qn3: idiag = ',idiag
    print*, 'm1qn3: iaux = ',iaux
    print*, 'm1qn3: ialpha = ',ialpha
    print*, 'm1qn3: iybar = ',iybar
    print*, 'm1qn3: isbar = ',isbar

    ! call the optimization code

 9999 continue

      call m1qn3a (n,x,f,g,dxmin,df1,epsg, &
                   normtype,impres,io,imode,omode,niter,nsim,inmemo, &
                   iz(3),iz(4),iz(5),dz(id),dz(igg),dz(idiag),dz(iaux), &
                   dz(ialpha),dz(iybar),dz(isbar),reverse,reentry,indic,dzs, &
                   sscale,cold,warm,i,itmax,moderl,isim,jcour,d1,t,tmin,tmax, &
                   gnorm,gnorms,eps1,ff,preco,precos,ys,den,dk,dk1,ps,ps2,hp0,&
                   t_increased,indica,indicd,&
                   tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd,&
                   fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa)

      if (reentry.gt.0) return

    ! impressions finales

 1000 continue
      if (impres.ge.1) write (io,960) omode,niter,nsim,epsg
  960 format (/1x,79("-")/ &
             /" m1qn3: output mode is ",i2 &
             /5x,"number of iterations: ",i14 &
             /5x,"number of simulations: ",i13 &
             /5x,"realized relative precision on g: ",1pd9.2)
      if (normtype.eq.'two') then
          gnorm = sqrt(dot_product(g,g))
      elseif (normtype.eq.'sup') then
          gnorm = dnrmi(n,g)
      elseif (normtype.eq.'dfn') then
          ps = dot_product(g,g)
          gnorm = dsqrt(ps)
      endif

      if (impres.ge.1) write (io,961) f,normtype,gnorm
  961 format (5x,"f             = ",1pd15.8 &
             /5x,a3,"-norm of g = ",1pd15.8)

      return

      end subroutine m1qn3


  !------------------------------------------------------------
  ! M1QN3A
  !------------------------------------------------------------

      subroutine m1qn3a (n,x,f,g,dxmin,df1,epsg, &
                         normtype,impres,io,imode,omode,niter,nsim,inmemo, &
                         m,jmin,jmax,d,gg,diag,aux, &
                         alpha,ybar,sbar,reverse,reentry,indic,dzs, &
                         sscale,cold,warm,i,itmax,moderl,isim,jcour,d1,t,tmin,tmax, &
                         gnorm,gnorms,eps1,ff,preco,precos,ys,den,dk,dk1,ps,ps2,hp0,&
                         t_increased,indica,indicd,&
                         tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd,&
                         fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa)

      implicit none

    ! arguments

      logical :: inmemo
      character(len=3) :: normtype
      integer :: n,impres,io,imode(3),omode,niter,nsim,m,jmin,jmax,indic,reverse,reentry
      real(kind=8) :: x(n),f,g(n),dxmin,df1,epsg,d(n),gg(n),diag(n),aux(n),alpha(m),ybar(n,m),sbar(n,m),dzs(*)

    ! variables locales

      logical :: sscale,cold,warm,skip_update
      integer :: i,itmax,moderl,isim,jcour
      real(kind=8) :: d1,t,tmin,tmax,gnorm,gnorms,eps1,ff,preco,precos,ys,den,dk,dk1,ps,ps2,hp0


    ! variables defined in mlis3 but need to be saved

      logical ::t_increased
      integer :: indica,indicd
      real(kind=8) :: tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd
      real(kind=8) :: fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa

    ! parametres

      real(kind=8), parameter :: rm1=0.0001d+0
      real(kind=8), parameter :: rm2=0.99d+0
      real(kind=8), parameter :: pi=3.1415927d+0
      real(kind=8) :: rmin

    ! possible jumps
    !     9998: call of the simulator in m1qn3a with indic = 1
    !     9999: call of the simulator in mlis3 with indic = 4

      print*, 'm1qn3a: reentry = ',reentry
      print*, 'm1qn3a: isim = ',isim
      print*, 'm1qn3a: normtype = ',normtype

      if (reentry.eq.1) goto 9998
      if (reentry.eq.2) goto 9999

    ! initialisation

      rmin=1.d-20

      sscale=.true.
      if (imode(1).eq.0) sscale=.false.

      warm=.false.
      if (imode(2).eq.1) warm=.true.
      cold=.not.warm

      skip_update = .false.

      itmax=niter
      niter=0
      eps1=1.d+0

      ps = dot_product(g,g)
      gnorm = dsqrt(ps)
      if (normtype.eq.'two') then
          gnorms = sqrt(dot_product(g,g))
      elseif (normtype.eq.'sup') then
          gnorms = dnrmi(n,g)
      elseif (normtype.eq.'dfn') then
          gnorms = gnorm
      endif
      if (impres.ge.1) write (io,900) f,normtype,gnorms
  900 format (5x,"f             = ",1pd15.8 &
             /5x,a3,"-norm of g = ",1pd15.8)
      if (gnorms.lt.rmin) then
          omode=2
          if (impres.ge.1) write (io,901)
          goto 1000
      endif
  901 format (/" >>> m1qn3a: initial gradient is too small")

    print*, 'm1qn3a: gnorms = ',gnorms

    ! initialisation pour dd

      if (cold) then
          jmin=1
          jmax=0
      endif
      jcour=1
      if (inmemo) jcour=jmax

    ! mise a l'echelle de la premiere direction de descente

      if (cold) then

    ! use Fletcher's scaling and initialize diag to 1.

          precos=2.d+0*df1/gnorm**2
          do i=1,n
              d(i)=-g(i)*precos
              diag(i)=1.d+0
          end do
          if (impres.ge.5) write(io,902) precos
  902     format (/" m1qn3a: descent direction -g: precon = ",d10.3)
      else

    ! use the matrix stored in [diag and] the (y,s) pairs

          if (sscale) then
              ps = dot_product(ybar(:,jcour),ybar(:,jcour))
              precos=1.d+0/ps
          endif
          do i=1,n
              d(i)=-g(i)
          end do
          if (inmemo) then
              call dd (n,sscale,m,d,aux,jmin,jmax, &
                       precos,diag,alpha,ybar,sbar,dzs)
          else
              call dds (n,sscale,m,d,aux,jmin,jmax, &
                        precos,diag,alpha,ybar,sbar,dzs)
          endif
      endif

      if (impres.eq.3) write(io,903)
      if (impres.eq.4) write(io,903)
  903 format (/1x,79("-"))
  904 format (1x)

    ! initialisation pour mlis3

      tmax=1.d+20
      hp0 = dot_product(d,g)
      if (hp0.ge.0.d+0) then
          omode=7
          if (impres.ge.1) write (io,905) niter,hp0
          goto 1000
      endif
  905 format (/" >>> m1qn3 (iteration ",i2,"): " &
              /5x," the search direction d is not a ", &
              "descent direction: (g,d) = ",d12.5)

    ! compute the angle (-g,d)

      if (warm.and.impres.ge.5) then
          ps = dot_product(g,g)
          ps = dsqrt(ps)
          ps2 = dot_product(d,d)
          ps2=dsqrt(ps2)
          ps=hp0/ps/ps2
          ps=dmin1(-ps,1.d+0)
          ps=dacos(ps)
          d1=ps*180.d+0/pi
          write (io,906) sngl(d1)
      endif
  906 format (/" m1qn3: descent direction d: ", &
              "angle(-g,d) = ",f5.1," degrees")

    ! Debut de l iteration. on cherche x(k+1) de la forme x(k) + t*d,
    ! avec t > 0. On connait d.

    ! Debut de la boucle: etiquette 100,
    ! Sortie de la boucle: goto 1000.

100   niter=niter+1
      if (impres.ge.5) write(io,903)
      if (impres.ge.4) write(io,904)
      if (impres.ge.4) write (io,910) niter,isim,f,hp0
  910 format (" m1qn3: iter ",i0,", simul ",i0, &
              ", f=",1pd15.8,", h'(0)=",d12.5)

    ! free simulation if desired

      if (imode(3).gt.0) then
          if (mod(niter-1,imode(3)).eq.0) then
              indic=1
              if (reverse.gt.0) then
                ! reverse communication
                ! exit and go to simul
                  reentry = 1
                  return
              else
                  call simul(indic,n,x,f,g,dzs)
              endif
          endif
      endif
 9998 continue

    ! recherche lineaire et nouveau point x(k+1)

      do i=1,n
          gg(i)=g(i)
      end do
      ff=f
      if (impres.ge.5) write (io,911)
  911 format (/" m1qn3: line search")

    ! calcul de tmin

      tmin=0.d+0
      do i=1,n
          tmin=dmax1(tmin,dabs(d(i)))
      end do
      tmin=dxmin/tmin
      t=1.d+0
      d1=hp0

 9999 continue
      call mlis3 (n,x,f,d1,t,tmin,tmax,d,g,rm2,rm1,impres, &
                  io,moderl,isim,nsim,aux,reverse,reentry,indic,dzs, &
                  t_increased,i,indica,indicd,tesf,tesd,tg,fg,fpg,td,ta,fa, &
                  fpa,d2,fn,fp,ffn,fd,fpd,z,test,barmin,barmul,barmax,barr, &
                  gauche,droite,taa,ps)

      if (reentry.gt.0) return

    ! mlis3 renvoie les nouvelles valeurs de x, f et g

      if (moderl.ne.0) then
          if (moderl.lt.0) then

    ! calcul impossible
    ! t, g: ou les calculs sont impossibles
    ! x, f: ceux du t_gauche (donc f <= ff)

              omode=moderl
              goto 1000
          elseif (moderl.eq.1) then

    ! descente bloquee sur tmax

              skip_update = .true.
!             omode=3
!             if (impres.ge.1) write(io,912) niter
! 912         format (/" >>> m1qn3 (iteration ",i0,
!    &                "): line search blocked on tmax: "/
!    &                " >>> possible reasons: bad scaling,",
!    &                " unbounded problem")
          elseif (moderl.eq.4) then

    ! nsim atteint
    ! x, f: ceux du t_gauche (donc f <= ff)

              omode=5
              goto 1000
          elseif (moderl.eq.5) then

    ! arret demande par l utilisateur (indic = 0)
    ! x, f: ceux en sortie du simulateur

              omode=0
              goto 1000
          elseif (moderl.eq.6) then

    ! arret sur dxmin ou appel incoherent
    ! x, f: ceux du t_gauche (donc f <= ff)

              omode=6
              goto 1000
          endif
      else
          skip_update = .false.
      endif

    ! NOTE: stopping tests are now done after having updated the matrix, so
    ! that update information can be stored in case of a later warm restart

    ! mise a jour de la matrice

      if (skip_update) then
          if (impres.ge.5) write(io,'(/a)') &
             " m1qn3: matrix update is skipped"
      elseif (m.gt.0) then

    ! mise a jour des pointeurs

          jmax=jmax+1
          if (jmax.gt.m) jmax=jmax-m
          if ((cold.and.niter.gt.m).or.(warm.and.jmin.eq.jmax)) then
              jmin=jmin+1
              if (jmin.gt.m) jmin=jmin-m
          endif
          if (inmemo) jcour=jmax

    ! y, s et (y,s)

          do i=1,n
              sbar(i,jcour)=t*d(i)
              ybar(i,jcour)=g(i)-gg(i)
          end do
          if (impres.ge.5) then
              ps = dot_product(sbar(:,jcour),sbar(:,jcour))
              dk1=dsqrt(ps)
              print*, 'm1qn3a: ps = ',ps
              print*, 'm1qn3a: dk, dk1 = ',dk, dk1
              if (niter.gt.1) write (io,930) dk1/dk
  930         format (/" m1qn3: convergence rate, s(k)/s(k-1) = ", &
                      1pd12.5)
              dk=dk1
          endif
          ys = dot_product(ybar(:,jcour),sbar(:,jcour))
          if (ys.le.0.d+0) then
              omode=7
              if (impres.ge.1) write (io,931) niter,ys
  931         format (/" >>> m1qn3 (iteration ",i2, &
                      "): the scalar product (y,s) = ",d12.5 &
                      /27x,"is not positive")
              goto 1000
          endif

    ! ybar et sbar

          d1=dsqrt(1.d+0/ys)
          do i=1,n
              sbar(i,jcour)=d1*sbar(i,jcour)
              ybar(i,jcour)=d1*ybar(i,jcour)
          end do
!          if (.not.inmemo) call ystbl (.true.,ybar,sbar,n,jmax)

    ! compute the scalar or diagonal preconditioner

          if (impres.ge.5) write(io,932)
  932     format (/" m1qn3: matrix update:")

    ! Here is the Oren-Spedicato factor, for scalar scaling

          if (sscale) then
              ps = dot_product(ybar(:,jcour),ybar(:,jcour))
              precos=1.d+0/ps

              if (impres.ge.5) write (io,933) precos
  933         format (5x,"Oren-Spedicato factor = ",d10.3)

    ! Scale the diagonal to Rayleigh s ellipsoid.
    ! Initially (niter.eq.1) and for a cold start, this is
    ! equivalent to an Oren-Spedicato scaling of the
    ! identity matrix.

          else
              aux = ybar(:,jcour)
              ps=0.d0
              do i=1,n
                  ps=ps+diag(i)*aux(i)*aux(i)
              end do
              d1=1.d0/ps
              if (impres.ge.5) then
                  write (io,934) d1
  934             format(5x,"fitting the ellipsoid: factor = ",1pd10.3)
              endif
              do i=1,n
                  diag(i)=diag(i)*d1
              end do

    ! update the diagonal
    ! (gg is used as an auxiliary vector)

              gg = sbar(:,jcour)
              ps=0.d0
              do i=1,n
                  ps=ps+gg(i)*gg(i)/diag(i)
              end do
              den=ps
              do i=1,n
                  diag(i)=1.d0/ &
                         (1.d0/diag(i)+aux(i)**2-(gg(i)/diag(i))**2/den)
                  if (diag(i).le.0.d0) then
                      if (impres.ge.5) write (io,935) i,diag(i),rmin
                      diag(i)=rmin
                  endif
              end do
  935         format (/" >>> m1qn3-WARNING: diagonal element ",i8, &
                       " is negative (",d10.3,"), reset to ",d10.3)

              if (impres.ge.5) then
                  ps=0.d0
                  do i=1,n
                      ps=ps+diag(i)
                  end do
                  ps=ps/n
                  preco=ps

                  ps2=0.d0
                  do i=1,n
                      ps2=ps2+(diag(i)-ps)**2
                  end do
                  ps2=dsqrt(ps2/n)
                  write (io,936) preco,ps2
  936             format (5x,"updated diagonal: average value = ",&
                          1pd10.3,", sqrt(variance) = ",d10.3)
              endif
          endif
      endif

    ! tests d arret

      ps = dot_product(g,g)
      if (normtype.eq.'two') then
          gnorm = sqrt(dot_product(g,g))
      elseif (normtype.eq.'sup') then
          gnorm = dnrmi(n,g)
      elseif (normtype.eq.'dfn') then
          gnorm = dsqrt(ps)
      endif
      eps1 = gnorm/gnorms

!      print*, 'm1qn3a: ybar = ',ybar
!      print*, 'm1qn3a: sbar = ',sbar
!      print*, 'm1qn3a: g = ',g
      print*, 'm1qn3a: ps = ',ps
      print*, 'm1qn3a: gnorm = ',gnorm
      print*, 'm1qn3a: gnorms = ',gnorms
      print*, 'm1qn3a: eps1 = ',eps1

      if (impres.eq.3) then
          if (mod(niter-1,50).eq.0) write(io,'(/a,a)') &
             "  iter  simul  stepsize            f                |g|", &
             "       |g|/|g0|"
          write(io, &
             '(1x,i5,2x,i5,2x,1pd8.2,2x,d21.14,2x,d11.5,2x,d10.4)') &
             niter, isim, t, f, gnorm, eps1
      endif
      if (impres.ge.5) write (io,940) eps1
  940 format (/" m1qn3: stopping criterion on g: ",1pd12.5)
      if (eps1.lt.epsg) then
          omode=1
          goto 1000
      endif
      if (niter.eq.itmax) then
          omode=4
          if (impres.ge.1) write (io,941) niter
  941     format (/" >>> m1qn3 (iteration ",i0, &
                  "): maximal number of iterations")
          goto 1000
      endif
      if (isim.gt.nsim) then
          omode=5
          if (impres.ge.1) write (io,942) niter,isim
  942     format (/" >>> m1qn3 (iteration ",i3,"): ",i6, &
                  " simulations (maximal number reached)")
          goto 1000
      endif

    ! calcul de la nouvelle direction de descente d = - H.g

      if (m.eq.0) then
          preco=2.d0*(ff-f)/ps
          do i=1,n
              d(i)=-g(i)*preco
          end do
      else
          do i=1,n
              d(i)=-g(i)
          end do
          if (inmemo) then
              call dd (n,sscale,m,d,aux,jmin,jmax, &
                       precos,diag,alpha,ybar,sbar,dzs)
          else
              call dds (n,sscale,m,d,aux,jmin,jmax, &
                        precos,diag,alpha,ybar,sbar,dzs)
          endif
      endif

    ! test: la direction d est-elle de descente ?
    !       hp0 sera utilise par mlis3

      hp0 = dot_product(d,g)
      if (hp0.ge.0.d+0) then
          omode=7
          if (impres.ge.1) write (io,905) niter,hp0
          goto 1000
      endif
      if (impres.ge.5) then
          ps = dot_product(g,g)
          ps=dsqrt(ps)
          ps2 = dot_product(d,d)
          ps2=dsqrt(ps2)
          ps=hp0/ps/ps2
          ps=dmin1(-ps,1.d+0)
          ps=dacos(ps)
          d1=ps
          d1=d1*180.d0/pi
          write (io,906) sngl(d1)
      endif

    ! on poursuit les iterations

      goto 100

   ! m1qn3 has finished for ever

 1000 continue
      if (reverse.ne.0) reverse = -1
      reentry = 0
      nsim=isim
      epsg=eps1
      return

      end subroutine m1qn3a

  !------------------------------------------------------------
  ! DD
  !------------------------------------------------------------
  !
  !   calcule le produit H.g ou
  !       . H est une matrice construite par la formule de bfgs inverse
  !         a nm memoires a partir de la matrice diagonale diag
  !         dans un espace hilbertien dont le produit scalaire
  !         est donne par prosca
  !         (cf. J. Nocedal, Math. of Comp. 35/151 (1980) 773-782)
  !       . g est un vecteur de dimension n (en general le gradient)
  !
  !   la matrice diag apparait donc comme un preconditionneur diagonal
  !
  !   depl = g (en entree), = H g (en sortie)
  !
  !   la matrice H est memorisee par les vecteurs des tableaux
  !   ybar, sbar et les pointeurs jmin, jmax
  !
  !   alpha(nm) est une zone de travail
  !
  !   izs(*),rzs(*),dzs(*) sont des zones de travail pour prosca
  !
  !------------------------------------------------------------

     subroutine dd (n,sscale,nm,depl,aux,jmin,jmax, &
                     precos,diag,alpha,ybar,sbar,dzs)

      implicit none

    ! arguments

      logical :: sscale
      integer :: n,m,nm,jmin,jmax 
      real(kind=8) :: depl(n),precos,diag(n),alpha(nm),ybar(n,nm),sbar(n,nm),aux(n),dzs(*)

    ! variables locales

      integer :: jfin,i,j,jp
      real(kind=8) :: r,ps

      jfin=jmax
      if (jfin.lt.jmin) jfin=jmax+nm

    ! phase de descente

      do j=jfin,jmin,-1
          jp=j
          if (jp.gt.nm) jp=jp-nm
          ps = dot_product(depl,sbar(:,jp))
          r=ps
          alpha(jp)=r
          do i=1,n
              depl(i)=depl(i)-r*ybar(i,jp)
          end do
      end do

   ! preconditionnement

      if (sscale) then
          do i=1,n
              depl(i)=depl(i)*precos
          end do
      else
          aux = depl
          do i=1,n
              aux(i)=aux(i)*diag(i)
          end do
           depl = aux
      endif

    ! remontee

      do j=jmin,jfin
          jp=j
          if (jp.gt.nm) jp=jp-nm
          ps = dot_product(depl,ybar(:,jp))
          r=alpha(jp)-ps
          do i=1,n
              depl(i)=depl(i)+r*sbar(i,jp)
          end do 
      end do
      return

      end subroutine dd

  !------------------------------------------------------------
  ! DDS
  !------------------------------------------------------------
  !
  !   This subroutine has the same role as dd (computation of the
  !   product H.g). It supposes however that the (y,s) pairs are not
  !   stored in core memory, but on a devise chosen by the user.
  !   The access to this devise is performed via the subroutine ystbl.
  !
  !------------------------------------------------------------

      subroutine dds (n,sscale,nm,depl,aux,jmin, &
                      jmax,precos,diag,alpha,ybar,sbar,dzs)

      implicit none

    ! arguments

      logical :: sscale
      integer :: n,nm,jmin,jmax 
      real(kind=8) :: depl(n),precos,diag(n),alpha(nm),ybar(n),sbar(n),aux(n),dzs(*)

    ! variables locales

      integer :: jfin,i,j,jp
      real(kind=8) :: r,ps

      jfin=jmax
      if (jfin.lt.jmin) jfin=jmax+nm

    ! phase de descente

      do j=jfin,jmin,-1
          jp=j
          if (jp.gt.nm) jp=jp-nm
          ps = dot_product(depl,sbar)
          r=ps
          alpha(jp)=r
          do i=1,n
              depl(i)=depl(i)-r*ybar(i)
          end do
      end do

    ! preconditionnement

      if (sscale) then
          do i=1,n
              depl(i)=depl(i)*precos
         end do
      else
          do i=1,n
              aux(i)=aux(i)*diag(i)
          end do
      endif

    ! remontee

      do j=jmin,jfin
          jp=j
          if (jp.gt.nm) jp=jp-nm
          ps = dot_product(depl,ybar)
          r=alpha(jp)-ps
          do i=1,n
              depl(i)=depl(i)+r*sbar(i)
          end do
      end do
      return

      end subroutine dds


  !------------------------------------------------------------
  ! MLIS3
  !------------------------------------------------------------
  !
  !   mlis3 + minuscules + commentaires
  !   + version amelioree (XII 88): interpolation cubique systematique
  !   et anti-overflows
  !   + declaration variables (II/89, JCG).
  !   + barr is also progressively decreased (12/93, CL & JChG).
  !   barmul is set to 5.
  !
  !-------------------------------------------------------------
  !
  !        en sortie logic =
  !
  !        0          descente serieuse
  !        1          descente bloquee sur tmax
  !        4          nap > napmax
  !        5          arret demande par le simulateur
  !        6          fonction et gradient pas d accord
  !        < 0        contrainte implicite active
  !
  !     indic on entry (in reverse communication)
  !
  !       < 0: the simulator cannot compute f and g
  !       = 0: the simulator wants to stop
  !       > 0: the simulator has done its work
  !
  !     indic on return (in reverse communication)
  !
  !       = 4: the simulator is asked to compute f and g
  !
  !     reverse
  !
  !       = 0: direct communication
  !       = 1: reverse communication
  !
  !     reentry on entry
  !
  !       = 2: reverse communication, return from a simulation, skip to
  !            the place where the simulator was called (9999)
  !
  !     reentry return
  !
  !       = 0: reverse communication, the linesearch has finished its job
  !       = 2: reverse communication, a simulation is required
  !
  !------------------------------------------------------------

      subroutine mlis3 (n,x,f,fpn,t,tmin,tmax,d,g, &
                        amd,amf,imp,io,logic,nap,napmax,xn, &
                        reverse,reentry,indic,dzs, &
                        t_increased,i,indica,indicd,tesf,tesd,tg,fg,fpg,td,ta,fa,&
                        fpa,d2,fn,fp,ffn,fd,fpd,z,test,barmin,barmul,barmax,barr,&
                        gauche,droite,taa,ps)

      implicit none

    ! arguments

      integer :: n,imp,io,logic,nap,napmax,indic,reverse,reentry
      real(kind=8) :: x(n),f,fpn,t,tmin,tmax,d(n),g(n),amd,amf,xn(n),dzs(*)

    ! variables locales

      logical ::t_increased
      integer :: i,indica,indicd
      real(kind=8) :: tesf,tesd,tg,fg,fpg,td,ta,fa,fpa,d2,fn,fp,ffn,fd
      real(kind=8) :: fpd,z,test,barmin,barmul,barmax,barr,gauche,droite,taa,ps

 1000 format (/4x," mlis3   ",4x,"fpn=",1pd10.3," d2=",d9.2,"  tmin=",d9.2," tmax=",d9.2)
 1001 format (/4x," mlis3",3x,"stop on tmin",5x,"stepsizes",11x,"functions",8x,"derivatives")
 1002 format (4x," mlis3",37x,1pd10.3,2d11.3)
 1003 format (4x," mlis3",1pd14.3,2d11.3)
 1004 format (4x," mlis3",37x,1pd10.3," indic=",i3)
 1005 format (4x," mlis3",14x,1pd18.8,2x,d21.14,2x,d11.4)
 1006 format (4x," mlis3",14x,1pd18.8,"      indic=",i3)
 1007 format (/4x," mlis3",10x,"tmin forced to tmax")
 1008 format (/4x," mlis3",10x,"inconsistent call")

    ! possible jump

      if (reentry.eq.2) goto 9999

      if (n.gt.0 .and. fpn.lt.0.d0 .and. t.gt.0.d0 &
      .and. tmax.gt.0.d0 .and. amf.gt.0.d0 &
      .and. amd.gt.amf .and. amd.lt.1.d0) go to 5
      logic=6
      go to 999
    5 tesf=amf*fpn
      tesd=amd*fpn
      barmin=0.01d0
      barmul=5.d0
      barmax=0.3d0
      barr=barmin
      td=0.d0
      tg=0.d0
      fn=f
      fg=fn
      fpg=fpn
      ta=0.d0
      fa=fn
      fpa=fpn
      ps = dot_product(d,d)
      d2=ps

    ! elimination d un t initial ridiculement petit
!     if (t.gt.tmin) go to 20
!     t=tmin
!     if (t.le.tmax) go to 20
!     if (imp.gt.0) write (io,1007)
!     tmin=tmax
!  20 if (fn+t*fpn.lt.fn+0.9d0*t*fpn) go to 30
!     t=2.d0*t
!     go to 20

    ! changed into
      if (t.lt.tmin) then
          t=tmin
          if (imp.ge.4) write (io,'(a)') &
             ' mlis3: initial step-size increased to tmin'
          if (t.gt.tmax) then
              if (imp.gt.0) write (io,1007)
              tmin=tmax
          endif
      endif
      t_increased = .false.
      do while (fn+t*fpn.ge.fn+0.9d0*t*fpn)
          t_increased = .true.
          t = 2.d0*t
      enddo
      if (t_increased .and. (imp.ge.4)) write (io,'(a,1pd10.3)') &
         ' mlis3: initial step-size increased to ',t

   30 indica=1
      logic=0
      if (t.gt.tmax) then
          t=tmax
          logic=1
      endif
      if (imp.ge.4) write (io,1000) fpn,d2,tmin,tmax

    ! nouveau x
    ! initialize xn to the current iterate
    ! use x as the trial iterate

      do i=1,n
          xn(i)=x(i)
          x(i)=xn(i)+t*d(i)
      enddo

    ! boucle

  100 nap=nap+1
      if(nap.gt.napmax) then
          logic=4
          fn=fg
          do i=1,n
              xn(i)=xn(i)+tg*d(i)
          enddo
          go to 999
      endif

    ! appel simulateur

      indic=4
      if (reverse.gt.0) then
          ! reverse communication
          ! exit and go to simul
          reentry = 2
          return
      else
          call simul(indic,n,x,f,g,dzs)
      endif
 9999 continue
      if (indic.eq.0) then

    ! arret demande par l utilisateur

          logic=5
          fn=f
          do i=1,n
              xn(i)=x(i)
          end do
          go to 999
      endif
      if(indic.lt.0) then

    ! les calculs n ont pas pu etre effectues par le simulateur

          td=t
          indicd=indic
          logic=0
          if (imp.ge.4) write (io,1004) t,indic
          t=tg+0.1d0*(td-tg)
          go to 905
      endif

    ! les tests elementaires sont faits, on y va

      ps = dot_product(d,g)
      fp=ps

    ! premier test de Wolfe

      ffn=f-fn
      print*, 'mlis3: 1st Wolfe test'
      print*, 'mlis3: ffn, t, tesf = ',ffn,t,tesf
      if(ffn.gt.t*tesf) then
          td=t
          fd=f
          fpd=fp
          indicd=indic
          logic=0
          if(imp.ge.4) write (io,1002) t,ffn,fp
          go to 500
      endif

    ! test 1 ok, donc deuxieme test de Wolfe

      print*, 'mlis3: 2nd Wolfe test'
      print*, 'mlis3: fp, tesd = ',fp,tesd
      if(imp.ge.4) write (io,1003) t,ffn,fp
      if(fp.gt.tesd) then
          logic=0
          go to 320
      endif
      if (logic.eq.0) go to 350

    ! test 2 ok, donc pas serieux, on sort

  320 fn=f
      do i=1,n
          xn(i)=x(i)
      end do
      go to 999


  350 tg=t
      fg=f
      fpg=fp
      if(td.ne.0.d0) go to 500

    !  extrapolation

      taa=t
      gauche=(1.d0+barmin)*t
      droite=10.d0*t
      call ecube (t,f,fp,ta,fa,fpa,gauche,droite)
      ta=taa
      if(t.lt.tmax) go to 900
      logic=1
      t=tmax
      go to 900

    ! interpolation

  500 if(indica.le.0) then
          ta=t
          t=0.9d0*tg+0.1d0*td
          go to 900
      endif
      test=barr*(td-tg)
      gauche=tg+test
      droite=td-test
      taa=t
      call ecube (t,f,fp,ta,fa,fpa,gauche,droite)
      ta=taa
      if (t.gt.gauche .and. t.lt.droite) then
          barr=dmax1(barmin,barr/barmul)
!         barr=barmin
        else
          barr=dmin1(barmul*barr,barmax)
      endif

    ! fin de boucle
    !   - t peut etre bloque sur tmax
    !     (venant de l extrapolation avec logic=1)

  900 fa=f
      fpa=fp
  905 indica=indic

    ! faut-il continuer ?

      print*, 'mlis3: td = ',td
      if (td.eq.0.d0) go to 950
      if (td-tg.lt.tmin) go to 920

    ! limite de precision machine (arret de secours) ?

      do i=1,n
          z=xn(i)+t*d(i)
          if (z.ne.xn(i).and.z.ne.x(i)) go to 950
      end do
      if (imp.gt.3) write (io,'(5x,a)') "mlis3   no change in x"

    ! arret sur dxmin ou de secours

  920 logic=6

    ! si indicd<0, derniers calculs non faits par simul

      if (indicd.lt.0) logic=indicd

    ! si tg=0, xn = xn_depart,
    ! sinon on prend xn=x_gauche qui fait decroitre f

      if (tg.eq.0.d0) go to 940
      fn=fg
      do i=1,n
          xn(i)=xn(i)+tg*d(i)
      end do
  940 if (imp.le.3) go to 999
      write (io,1001)
      write (io,1005) tg,fg,fpg
      if (logic.eq.6) write (io,1005) td,fd,fpd
      if (logic.eq.7) write (io,1006) td,indicd
      go to 999

    ! recopiage de x et boucle

  950 do i=1,n
          x(i)=xn(i)+t*d(i)
      end do
      go to 100
    ! linesearch finished, no skip at next entry in mlis3
  999 if (reverse.ne.0) reentry = 0
      print*, 'mlis3: reentry = ',reentry
      return

      end subroutine mlis3


  !------------------------------------------------------------
  ! ECUBE
  !------------------------------------------------------------

      subroutine ecube (t,f,fp,ta,fa,fpa,tlower,tupper)

      implicit none

    ! arguments

      real(kind=8) :: sign,den,anum,t,f,fp,ta,fa,fpa,tlower,tupper

    ! variables locales

      real(kind=8) :: z1,b,discri

    ! Using f and fp at t and ta, computes new t by cubic formula
    ! safeguarded inside [tlower,tupper].

      z1=fp+fpa-3.d0*(fa-f)/(ta-t)
      b=z1+fp

    ! first compute the discriminant (without overflow)

      if (dabs(z1).le.1.d0) then
          discri=z1*z1-fp*fpa
        else
          discri=fp/z1
          discri=discri*fpa
          discri=z1-discri
          if (z1.ge.0.d0 .and. discri.ge.0.d0) then
              discri=dsqrt(z1)*dsqrt(discri)
              go to 120
          endif
          if (z1.le.0.d0 .and. discri.le.0.d0) then
              discri=dsqrt(-z1)*dsqrt(-discri)
              go to 120
          endif
          discri=-1.d0
      endif
      if (discri.lt.0.d0) then
          if (fp.lt.0.d0) t=tupper
          if (fp.ge.0.d0) t=tlower
          go to 900
      endif

   !  discriminant nonnegative, compute solution (without overflow)

      discri=dsqrt(discri)
  120 if (t-ta.lt.0.d0) discri=-discri
      sign=(t-ta)/dabs(t-ta)
      if (b*sign.gt.0.d+0) then
          t=t+fp*(ta-t)/(b+discri)
        else
          den=z1+b+fpa
          anum=b-discri
          if (dabs((t-ta)*anum).lt.(tupper-tlower)*dabs(den)) then
              t=t+anum*(ta-t)/den
            else
              t=tupper
          endif
      endif
  900 t=dmax1(t,tlower)
      t=dmin1(t,tupper)
      return

      end subroutine ecube

  !------------------------------------------------------------
  ! MUPDTS
  !------------------------------------------------------------
  !   On entry:
  !     sscale: .true. if scalar initial scaling,
  !             .false. if diagonal initial scaling
  !     n:      number of variables
  !
  !   This routine has to return:
  !     m:      the number of updates to form the approximate Hessien H,
  !     inmemo: .true., if the vectors y and s used to form H are stored
  !                in core memory,
  !             .false. otherwise (storage of y and s on disk, for
  !                instance).
  !   When inmemo=.false., the routine "ystbl", which stores and
  !   restores (y,s) pairs, has to be rewritten.
  !------------------------------------------------------------

      subroutine mupdts (sscale,inmemo,n,m,nrz)

      implicit none

    ! arguments

      logical :: sscale,inmemo
      integer :: n,m,nrz

      if (sscale) then
          m=(nrz-3*n)/(2*n+1)
      else
          m=(nrz-4*n)/(2*n+1)
      endif
      inmemo=.true.
      return

      end subroutine mupdts


  !------------------------------------------------------------
  ! SIMUL
  !------------------------------------------------------------
  ! Routine for the simulator if using direct communication
  ! 
  !------------------------------------------------------------

      subroutine simul (indic,n,x,f,g,dzs)

      implicit none
      integer :: indic,n
      real(kind=8) :: x(n),f,g(n),dzs(*)

    ! write simul routine here for direct communication

      return

      end subroutine simul

  !------------------------------------------------------------
  ! DNRMI
  !------------------------------------------------------------

      function dnrmi (n,v)

      integer :: n
      real(kind=8) :: v(n)
    ! Computes the infinty-norm of the vector v(n)
    ! local variables

      integer :: i
      real(kind=8) :: norm
      real(kind=8) :: dnrmi

    ! compute

      norm = 0.d0
      if (n.gt.0) then
        do i=1,n
           norm = max(norm,abs(v(i)))
        enddo
      endif
      dnrmi = norm
      return

      end function dnrmi

  !------------------------------------------------------------

end module mod_m1qn3

